Device Tree channel nodes are associated with the adapters created by
i2c-mux, but equivalent software-node descriptions are not.

Find the software-node child whose reg value matches the channel and
assign it to the new adapter. Keep a reference for the adapter lifetime
so downstream clients can find the channel by firmware node.

Signed-off-by: Ahmad Byagowi <[email protected]>
---
 drivers/i2c/i2c-mux.c | 69 +++++++++++++++++++++++++++++++++++++++++--
 1 file changed, 67 insertions(+), 2 deletions(-)

diff --git a/drivers/i2c/i2c-mux.c b/drivers/i2c/i2c-mux.c
index 681a201c239b..8b54bb13c613 100644
--- a/drivers/i2c/i2c-mux.c
+++ b/drivers/i2c/i2c-mux.c
@@ -20,11 +20,13 @@
  */
 
 #include <linux/acpi.h>
+#include <linux/err.h>
 #include <linux/i2c.h>
 #include <linux/i2c-mux.h>
 #include <linux/kernel.h>
 #include <linux/module.h>
 #include <linux/of.h>
+#include <linux/property.h>
 #include <linux/slab.h>
 #include <linux/sysfs.h>
 
@@ -264,6 +266,55 @@ static const struct i2c_lock_operations 
i2c_parent_lock_ops = {
        .unlock_bus =  i2c_parent_unlock_bus,
 };
 
+static struct fwnode_handle *
+i2c_mux_get_channel_swnode(struct i2c_mux_core *muxc, u32 chan_id)
+{
+       struct fwnode_handle *dev_node = dev_fwnode(muxc->dev);
+       struct fwnode_handle *mux_node, *child = NULL;
+       u32 reg;
+
+       /* A software node supplementing ACPI is the secondary fwnode. */
+       if (!is_software_node(dev_node)) {
+               if (IS_ERR_OR_NULL(dev_node))
+                       return NULL;
+               dev_node = dev_node->secondary;
+       }
+       if (!is_software_node(dev_node))
+               return NULL;
+
+       if (muxc->arbitrator)
+               mux_node = fwnode_get_named_child_node(dev_node, "i2c-arb");
+       else if (muxc->gate)
+               mux_node = fwnode_get_named_child_node(dev_node, "i2c-gate");
+       else
+               mux_node = fwnode_get_named_child_node(dev_node, "i2c-mux");
+
+       if (mux_node) {
+               /* A "reg" property indicates an old-style firmware entry. */
+               if (!fwnode_property_read_u32(mux_node, "reg", &reg)) {
+                       fwnode_handle_put(mux_node);
+                       mux_node = NULL;
+               }
+       }
+
+       if (!mux_node)
+               mux_node = fwnode_handle_get(dev_node);
+       else if (muxc->arbitrator || muxc->gate)
+               child = fwnode_handle_get(mux_node);
+
+       if (!child) {
+               fwnode_for_each_child_node(mux_node, child) {
+                       if (fwnode_property_read_u32(child, "reg", &reg))
+                               continue;
+                       if (chan_id == reg)
+                               break;
+               }
+       }
+
+       fwnode_handle_put(mux_node);
+       return child;
+}
+
 int i2c_mux_add_adapter(struct i2c_mux_core *muxc,
                        u32 force_nr, u32 chan_id)
 {
@@ -324,8 +375,8 @@ int i2c_mux_add_adapter(struct i2c_mux_core *muxc,
                priv->adap.lock_ops = &i2c_parent_lock_ops;
 
        /*
         * Try to populate the mux adapter's of_node, expands to
         * nothing if !CONFIG_OF.
         */
        if (muxc->dev->of_node) {
                struct device_node *dev_node = muxc->dev->of_node;
@@ -364,6 +415,18 @@ int i2c_mux_add_adapter(struct i2c_mux_core *muxc,
 
                priv->adap.dev.of_node = child;
                of_node_put(mux_node);
+       } else {
+               struct fwnode_handle *fwnode;
+               const struct software_node *swnode;
+
+               fwnode = i2c_mux_get_channel_swnode(muxc, chan_id);
+               if (fwnode) {
+                       swnode = to_software_node(fwnode);
+                       ret = device_add_software_node(&priv->adap.dev, swnode);
+                       fwnode_handle_put(fwnode);
+                       if (ret)
+                               goto err_free_priv;
+               }
        }
 
        /*
@@ -408,6 +471,7 @@ int i2c_mux_add_adapter(struct i2c_mux_core *muxc,
        return 0;
 
 err_free_priv:
+       device_remove_software_node(&priv->adap.dev);
        kfree(priv);
        return ret;
 }
@@ -429,6 +493,7 @@ void i2c_mux_del_adapters(struct i2c_mux_core *muxc)
                sysfs_remove_link(&muxc->dev->kobj, symlink_name);
 
                sysfs_remove_link(&priv->adap.dev.kobj, "mux_device");
+               device_remove_software_node(&priv->adap.dev);
                i2c_del_adapter(adap);
                of_node_put(np);
                kfree(priv);
@@ -324,8 +375,8 @@ int i2c_mux_add_adapter(struct i2c_mux_core *muxc,
 
        /*
-        * Try to populate the mux adapter's of_node, expands to
-        * nothing if !CONFIG_OF.
+        * Associate the mux adapter with its DT or software-node channel.
+        * DT support expands to nothing if !CONFIG_OF.
         */
        if (muxc->dev->of_node) {
-- 
2.50.1 (Apple Git-155)


Reply via email to